@@ -1,6 +1,4 @@
 /*
- * $Id: globals.h,v 2.44 1996/10/15 20:16:35 hzoli Exp $
- *
  * globals.h - global variables for zsh
  *
  * This file is part of zsh, the Z shell.
@@ -320,19 +318,19 @@ EXTERN char *cached_username;   /* $USERNAME   */
 EXTERN char *zsh_name;		/* ZSH_NAME    */
 
 EXTERN char *underscore;	/* $_          */
-EXTERN long lastval;            /* $?          */
-EXTERN long mypid;		/* $$          */
-EXTERN long lastpid;		/* $!          */
-EXTERN long ppid;		/* $PPID       */
+EXTERN zlong lastval;		/* $?          */
+EXTERN zlong mypid;		/* $$          */
+EXTERN zlong lastpid;		/* $!          */
+EXTERN zlong ppid;		/* $PPID       */
 EXTERN char *ifs;		/* $IFS        */
 EXTERN char *pwd;		/* $PWD        */
 EXTERN char *oldpwd;		/* $OLDPWD     */
 
-EXTERN long columns;            /* $COLUMNS    */
-EXTERN long lines;              /* $LINES      */
+EXTERN zlong columns;		/* $COLUMNS    */
+EXTERN zlong lines;		/* $LINES      */
 
 EXTERN char *zoptarg;		/* $OPTARG     */
-EXTERN long zoptind;		/* $OPTIND     */
+EXTERN zlong zoptind;		/* $OPTIND     */
 EXTERN char *prompt;		/* $PROMPT     */
 EXTERN char *prompt2;		/* etc.        */
 EXTERN char *prompt3;
@@ -355,8 +353,8 @@ EXTERN char *argzero;           /* $0 */
 EXTERN char *hackzero;
 EXTERN char *scriptname;        /* name of script being sourced */
 
-EXTERN long lineno;             /* $LINENO       */
-EXTERN long shlvl;              /* $SHLVL        */
+EXTERN zlong lineno;		/* $LINENO       */
+EXTERN zlong shlvl;		/* $SHLVL        */
  
 EXTERN long lastval2;
 
@@ -435,6 +433,10 @@ EXTERN FILE *shout;
 
 EXTERN FILE *bshin;
 
+/* the FILE for xtrace output */
+
+EXTERN FILE *xtrerr;
+
 /* != 0 means we are reading input from a string */
  
 EXTERN int strin;
@@ -534,6 +536,14 @@ EXTERN int useheap;
 EXTERN int clwsize, clwnum, clwpos;
 EXTERN char **clwords;
 
+/* Non-zero if a completion list was displayed. */
+
+EXTERN int listshown;
+
+/* Non-zero if refresh() should clear the list below the prompt. */
+
+EXTERN int clearlist;
+
 /* pid of process undergoing 'process substitution' */
  
 EXTERN pid_t cmdoutpid;
@@ -567,6 +577,11 @@ EXTERN int signal_queue[MAX_QUEUE_SIZE];
 EXTERN int queue_front;
 EXTERN int queue_rear;
 
+/* Previous values of errflag and breaks if the signal handler had to
+ * change them. And a flag saying if it did that. */
+
+EXTERN int prev_errflag, prev_breaks, errbrk_saved;
+
 /* 1 if aliases should not be expanded */
  
 EXTERN int noaliases;
@@ -614,7 +629,7 @@ char *tccapnams[TC_COUNT] =
 {
     "cl", "le", "LE", "nd", "RI", "up", "UP", "do",
     "DO", "dc", "DC", "ic", "IC", "cd", "ce", "al", "dl", "ta",
-    "md", "so", "us", "me", "se", "ue"
+    "md", "so", "us", "me", "se", "ue", "ch"
 };
 #else
 extern char *tccapnams[TC_COUNT];
@@ -637,7 +652,7 @@ char *tokstrings[WHILE + 1] = {
     ")",	/* OUTPAR	     */
     "||",	/* DBAR		     */
     "&&",	/* DAMPER	     */
-    ")",	/* OUTANG	  10 */
+    ">",	/* OUTANG	  10 */
     ">|",	/* OUTANGBANG	     */
     ">>",	/* DOUTANG	     */
     ">>|",	/* DOUTANGBANG	     */
@@ -685,11 +700,11 @@ extern struct option optns[OPT_SIZE];
 struct option optns[OPT_SIZE] = {
 # define x OPT_REV|
     {NULL, 0, 0, 0},
-    {"allexport", 		'a',  'a',  0},
+    {"allexport", 		'a',  'a',  OPT_EMULATE},
     {"alwayslastprompt", 	0,    0,    0},
     {"alwaystoend", 		0,    0,    0},
     {"appendhistory", 		0,    0,    0},
-    {"autocd", 			'J',  0,    0},
+    {"autocd", 			'J',  0,    OPT_EMULATE},
     {"autolist", 		'9',  0,    0},
     {"automenu", 		0,    0,    0},
     {"autonamedirs", 		0,    0,    0},
@@ -699,14 +714,14 @@ struct option optns[OPT_SIZE] = {
     {"autoremoveslash", 	0,    0,    0},
     {"autoresume", 		'W',  0,    0},
     {"badpattern", 		x'2', 0,    OPT_EMULATE|OPT_NONBOURNE},
-    {"banghist", 		x'K', 0,    OPT_EMULATE|OPT_NONBOURNE},
+    {"banghist", 		x'K', 0,    OPT_NONBOURNE},
     {"beep", 			x'B', 0,    OPT_ALL},
     {"bgnice", 			'6',  0,    OPT_EMULATE|OPT_NONBOURNE},
-    {"braceccl", 		0,    0,    0},
+    {"braceccl", 		0,    0,    OPT_EMULATE},
     {"bsdecho", 		0,    0,    OPT_EMULATE|OPT_SH},
-    {"cdablevars", 		'T',  0,    0},
-    {"chaselinks", 		'w',  0,    0},
-    {"clobber", 		x'C', x'C', OPT_ALL},
+    {"cdablevars", 		'T',  0,    OPT_EMULATE},
+    {"chaselinks", 		'w',  0,    OPT_EMULATE},
+    {"clobber", 		x'C', x'C', OPT_EMULATE|OPT_ALL},
     {"completealiases", 	0,    0,    0},
     {"completeinword", 		0,    0,    0},
     {"correct", 		'0',  0,    0},
@@ -716,16 +731,16 @@ struct option optns[OPT_SIZE] = {
     {"cshjunkiequotes", 	0,    0,    OPT_EMULATE|OPT_CSH},
     {"cshnullglob", 		0,    0,    OPT_EMULATE|OPT_CSH},
     {"equals", 			0,    0,    OPT_EMULATE|OPT_ZSH},
-    {"errexit", 		'e',  'e',  0},
+    {"errexit", 		'e',  'e',  OPT_EMULATE},
     {"exec", 			x'n', x'n', OPT_ALL},
-    {"extendedglob", 		0,    0,    0},
-    {"extendedhistory", 	0,    0,    OPT_EMULATE|OPT_CSH},
+    {"extendedglob", 		0,    0,    OPT_EMULATE},
+    {"extendedhistory", 	0,    0,    OPT_CSH},
     {"flowcontrol", 		0,    0,    OPT_ALL},
     {"functionargzero",		0,    0,    OPT_EMULATE|OPT_NONBOURNE},
-    {"glob", 			x'F', x'f', OPT_ALL},
+    {"glob", 			x'F', x'f', OPT_EMULATE|OPT_ALL},
     {"globassign", 		0,    0,    OPT_EMULATE|OPT_CSH},
     {"globcomplete", 		0,    0,    0},
-    {"globdots", 		'4',  0,    0},
+    {"globdots", 		'4',  0,    OPT_EMULATE},
     {"globsubst", 		0,    0,    OPT_EMULATE|OPT_NONZSH},
     {"hashcmds", 		0,    0,    OPT_ALL},
     {"hashdirs", 		0,    0,    OPT_ALL},
@@ -735,12 +750,13 @@ struct option optns[OPT_SIZE] = {
     {"histignoredups", 		'h',  0,    0},
     {"histignorespace", 	'g',  0,    0},
     {"histnostore", 		0,    0,    0},
+    {"histreduceblanks",	0,    0,    0},
     {"histverify", 		0,    0,    0},
     {"hup", 			0,    0,    OPT_EMULATE|OPT_ZSH},
     {"ignorebraces", 		'I',  0,    OPT_EMULATE|OPT_SH},
     {"ignoreeof", 		'7',  0,    0},
     {"interactive", 		'i',  'i',  OPT_SPECIAL},
-    {"interactivecomments", 	'k',  0,    OPT_EMULATE|OPT_BOURNE},
+    {"interactivecomments", 	'k',  0,    OPT_BOURNE},
     {"ksharrays", 		0,    0,    OPT_EMULATE|OPT_BOURNE},
     {"kshoptionprint", 		0,    0,    OPT_EMULATE|OPT_KSH},
     {"listambiguous", 		0,    0,    0},
@@ -749,7 +765,7 @@ struct option optns[OPT_SIZE] = {
     {"localoptions", 		0,    0,    OPT_EMULATE|OPT_KSH},
     {"login", 			'l',  'l',  OPT_SPECIAL},
     {"longlistjobs", 		'R',  0,    0},
-    {"magicequalsubst", 	0,    0,    0},
+    {"magicequalsubst", 	0,    0,    OPT_EMULATE},
     {"mailwarning", 		'U',  0,    0},
     {"markdirs", 		'8',  'X',  0},
     {"menucomplete", 		'Y',  0,    0},
@@ -758,20 +774,21 @@ struct option optns[OPT_SIZE] = {
     {"nomatch", 		x'3', 0,    OPT_EMULATE|OPT_NONBOURNE},
     {"notify", 			'5',  'b',  OPT_ZSH},
     {"nullglob", 		'G',  0,    OPT_EMULATE},
-    {"numericglobsort", 	0,    0,    0},
+    {"numericglobsort", 	0,    0,    OPT_EMULATE},
     {"overstrike", 		0,    0,    0},
-    {"pathdirs", 		'Q',  0,    0},
+    {"pathdirs", 		'Q',  0,    OPT_EMULATE},
     {"posixbuiltins",		0,    0,    OPT_EMULATE|OPT_BOURNE},
+    {"printeightbit", 		0,    0,    0},
     {"printexitvalue", 		'1',  0,    0},
     {"privileged", 		'p',  'p',  OPT_SPECIAL},
     {"promptcr", 		x'V', 0,    OPT_ALL},
-    {"promptsubst", 		0,    0,    OPT_EMULATE|OPT_KSH},
-    {"pushdignoredups", 	0,    0,    0},
-    {"pushdminus", 		0,    0,    0},
+    {"promptsubst", 		0,    0,    OPT_KSH},
+    {"pushdignoredups", 	0,    0,    OPT_EMULATE},
+    {"pushdminus", 		0,    0,    OPT_EMULATE},
     {"pushdsilent", 		'E',  0,    0},
-    {"pushdtohome", 		'D',  0,    0},
-    {"rcexpandparam", 		'P',  0,    0},
-    {"rcquotes", 		0,    0,    0},
+    {"pushdtohome", 		'D',  0,    OPT_EMULATE},
+    {"rcexpandparam", 		'P',  0,    OPT_EMULATE},
+    {"rcquotes", 		0,    0,    OPT_EMULATE},
     {"rcs", 			x'f', 0,    OPT_ALL},
     {"recexact", 		'S',  0,    0},
     {"rmstarsilent", 		'H',  0,    OPT_BOURNE},
@@ -779,7 +796,7 @@ struct option optns[OPT_SIZE] = {
     {"shglob", 			0,    0,    OPT_EMULATE|OPT_BOURNE},
     {"shinstdin", 		's',  's',  OPT_SPECIAL},
     {"shoptionletters",		0,    0,    OPT_EMULATE|OPT_BOURNE},
-    {"shortloops", 		0,    0,    OPT_ALL},
+    {"shortloops", 		0,    0,    OPT_NONBOURNE},
     {"shwordsplit", 		'y',  0,    OPT_EMULATE|OPT_BOURNE},
     {"singlecommand",		't',  't',  OPT_SPECIAL},
     {"singlelinezle", 		'M',  0,    OPT_KSH},
